import struct
from typing import Any, Callable, Generic, Optional, TypeVar, Union, cast, overload

from arcadeutils import FileBytes
from naomi.rom import NaomiEEPROMDefaults


class NaomiEEPRomException(Exception):
    pass


BytesLike = TypeVar("BytesLike", bound=Union[bytes, FileBytes])


class ArrayBridge:
    def __init__(self, parent: "NaomiEEPRom[Any]", valid_callback: Callable[[Union[bytes, FileBytes]], bool], name: str, length: int, offset1: int, offset2: int) -> None:
        self.name = name
        self.__callback = valid_callback
        self.__length = length
        self.__offset1 = offset1
        self.__offset2 = offset2
        self.__parent = parent

    @property
    def valid(self) -> bool:
        return self.__callback(self.__parent._data)

    @property
    def data(self) -> bytes:
        return self.__parent._data[self.__offset1:(self.__offset1 + self.__length)]

    @data.setter
    def data(self, databytes: bytes) -> None:
        self.__setitem__(slice(0, self.__length, 1), databytes)

    @overload
    def __getitem__(self, key: int) -> int:
        ...

    @overload
    def __getitem__(self, key: slice) -> bytes:
        ...

    def __getitem__(self, key: Union[int, slice]) -> Union[int, bytes]:
        if isinstance(key, int):
            if key < 0 or key >= self.__length:
                raise NaomiEEPRomException(f"Cannot get bytes outside of {self.name} EEPRom section!")

            # Arbitrarily choose the first section.
            realkey = key + self.__offset1
            return self.__parent._data[realkey]

        elif isinstance(key, slice):
            # Determine start of slice
            if key.start is None:
                start = self.__offset1
            elif key.start < 0:
                raise Exception("Do not support negative indexing!")
            else:
                start = self.__offset1 + key.start

            if key.stop is None:
                stop = self.__offset1 + self.__length
            elif key.stop < 0:
                raise Exception("Do not support negative indexing!")
            elif key.stop > self.__length:
                raise NaomiEEPRomException(f"Cannot get bytes outside of {self.name} EEPRom section!")
            else:
                stop = self.__offset1 + key.stop

            if key.step not in {None, 1}:
                raise NaomiEEPRomException(f"Cannot get bytes with stride != 1 for {self.name} EEPRom section!")

            # Arbitrarily choose the first section.
            return self.__parent._data[start:stop]

        else:
            raise NotImplementedError("Not implemented!")

    @overload
    def __setitem__(self, key: int, val: int) -> None:
        ...

    @overload
    def __setitem__(self, key: slice, val: bytes) -> None:
        ...

    def __setitem__(self, key: Union[int, slice], val: Union[int, bytes]) -> None:
        if isinstance(key, int):
            if key < 0 or key >= self.__length:
                raise NaomiEEPRomException(f"Cannot set bytes outside of {self.name} EEPRom section!")
            if isinstance(val, int):
                val = bytes([val])

            if len(val) != 1:
                raise NaomiEEPRomException(f"Cannot change size of {self.name} EEPRom section when assigning!")

            # Make sure we set both bytes in each section.
            for realkey in [key + self.__offset1, key + self.__offset2]:
                if isinstance(self.__parent._data, bytes):
                    self.__parent._data = self.__parent._data[:realkey] + val + self.__parent._data[realkey + 1:]
                elif isinstance(self.__parent._data, FileBytes):
                    self.__parent._data[realkey:(realkey + 1)] = val
                else:
                    raise Exception("Logic error!")

            # Sanity check what we did here.
            if len(self.__parent._data) != 128:
                raise Exception("Logic error!")

        else:
            # Determine start of slice
            if key.start is None:
                start = 0
            elif key.start < 0:
                raise Exception("Do not support negative indexing!")
            else:
                start = key.start

            if key.stop is None:
                stop = self.__length
            elif key.stop < 0:
                raise Exception("Do not support negative indexing!")
            elif key.stop > self.__length:
                raise NaomiEEPRomException(f"Cannot set bytes outside of {self.name} EEPRom section!")
            else:
                stop = key.stop

            if key.step not in {None, 1}:
                raise NaomiEEPRomException(f"Cannot set bytes with stride != 1 for {self.name} EEPRom section!")

            # Check input to function.
            if not isinstance(val, bytes):
                raise NaomiEEPRomException(f"Cannot change size of {self.name} EEPRom section when assigning!")
            if len(val) != (stop - start):
                raise NaomiEEPRomException(f"Cannot change size of {self.name} EEPRom section when assigning!")

            # Make sure we set both bytes in each section.
            for realstart, realstop in [
                (start + self.__offset1, stop + self.__offset1),
                (start + self.__offset2, stop + self.__offset2),
            ]:
                if isinstance(self.__parent._data, bytes):
                    self.__parent._data = self.__parent._data[:realstart] + val + self.__parent._data[realstop:]
                elif isinstance(self.__parent._data, FileBytes):
                    self.__parent._data[realstart:realstop] = val
                else:
                    raise Exception("Logic error!")

            if len(self.__parent._data) != 128:
                raise Exception("Logic error!")


class NaomiEEPRom(Generic[BytesLike]):

    @staticmethod
    def default(serial: bytes, system_defaults: Optional[NaomiEEPROMDefaults] = None, game_defaults: Optional[bytes] = None) -> "NaomiEEPRom[bytes]":
        if len(serial) != 4:
            raise NaomiEEPRomException("Invalid game serial!")

        # First, set up game defaults section.
        if game_defaults is None:
            game_settings = b'\xFF' * (128 - (18 * 2))
        else:
            header = NaomiEEPRom.crc(game_defaults) + struct.pack("<BB", len(game_defaults), len(game_defaults))
            game_settings = header + header + game_defaults + game_defaults

            padding_len = (128 - (18 * 2) - len(game_settings))
            if padding_len > 0:
                game_settings += b'\xFF' * padding_len

        # Now, set up the system defaults section.
        system_array = [0x10, *[s for s in serial], 0x09, 0x10, 0x00, 0x01, 0x01, 0x01, 0x00, 0x11, 0x11, 0x11, 0x11]
        if system_defaults is not None:
            # Only want to apply the defaults if the default section we were given requests it.
            if system_defaults.apply_settings:
                if system_defaults.force_vertical:
                    # Bottom half of byte.
                    system_array[0] = (system_array[0] & 0xF0) | 0x01
                if system_defaults.force_silent:
                    # Top half of byte.
                    system_array[0] = (system_array[0] & 0x0F) | 0x00
                if system_defaults.chute == "individual":
                    # Bottom half of byte.
                    system_array[6] = (system_array[6] & 0xF0) | 0x01
                if system_defaults.coin_setting > 0 and system_defaults.coin_setting <= 28:
                    # Whole byte, off by one.
                    system_array[7] = system_defaults.coin_setting - 1
                if system_defaults.coin_setting == 28:
                    # Gotta set individual assignments.
                    system_array[8] = system_defaults.credit_rate
                    system_array[9] = system_defaults.coin_1_rate
                    system_array[10] = system_defaults.coin_2_rate
                    system_array[11] = system_defaults.bonus

                # TODO: Skipping out looking up sequence texts for now. Tecchnically we should be setting the upper
                # and lower nibbles of the last 4 bytes to the system default sequence text values 1-8, but I was lazy.

        # Finally, construct the full EEPROM.
        system = bytes(system_array)
        system_crc = NaomiEEPRom.crc(system)
        return NaomiEEPRom(system_crc + system + system_crc + system + game_settings)

    def __init__(self, data: BytesLike) -> None:
        if len(data) != 128:
            raise NaomiEEPRomException("Invalid EEPROM length!")
        if not self.validate(data):
            raise NaomiEEPRomException("Invalid EEPROM CRC!")

        self._data: Union[bytes, FileBytes]
        if isinstance(data, FileBytes):
            self._data = data.clone()
        elif isinstance(data, bytes):
            self._data = data

    @staticmethod
    def __cap_32(val: int) -> int:
        return val & 0xFFFFFFFF

    @staticmethod
    def __crc_inner(running_crc: int, next_byte: int) -> int:
        # First, mask off the values so we don't get a collision
        running_crc &= 0xFFFFFF00
        next_byte &= 0xFF

        # Add the byte into the CRC
        running_crc = NaomiEEPRom.__cap_32(running_crc + next_byte)

        for _ in range(8):
            if running_crc < 0x80000000:
                running_crc = NaomiEEPRom.__cap_32(running_crc * 2)
            else:
                running_crc = NaomiEEPRom.__cap_32(running_crc * 2)
                running_crc = NaomiEEPRom.__cap_32(running_crc + 0x10210000)

        return running_crc

    @staticmethod
    def crc(data: bytes) -> bytes:
        running_crc = 0xDEBDEB00

        for byte in b"".join([data, b"\x00"]):
            running_crc = NaomiEEPRom.__crc_inner(running_crc, byte)

        final_crc = (running_crc >> 16) & 0xFFFF
        return struct.pack("<H", final_crc)

    @staticmethod
    def validate(data: Union[bytes, FileBytes], *, serial: Optional[bytes] = None) -> bool:
        # First, make sure its the right length.
        if len(data) != 128:
            return False

        # Now, before even checking CRCs (slow), check the serial section.
        if serial is not None and data[3:7] != serial:
            return False

        if not NaomiEEPRom.__validate_system(data):
            return False
        return NaomiEEPRom.__validate_game(data, blank_is_valid=True)

    @staticmethod
    def __validate_system(data: Union[bytes, FileBytes]) -> bool:
        # Returns whether the settings chunk passes CRC.
        sys_section1 = data[2:18]
        sys_section2 = data[20:36]

        if data[0:2] != NaomiEEPRom.crc(sys_section1):
            # The CRC doesn't match!
            return False
        if data[18:20] != NaomiEEPRom.crc(sys_section2):
            # The CRC doesn't match!
            return False

        # System section is good
        return True

    @staticmethod
    def __validate_game(data: Union[bytes, FileBytes], *, blank_is_valid: bool = False) -> bool:
        game_size1, game_size2 = struct.unpack("<BB", data[38:40])
        if game_size1 != game_size2:
            # These numbers should always match.
            return False

        game_size3, game_size4 = struct.unpack("<BB", data[42:44])
        if game_size3 != game_size4:
            # These numbers should always match.
            return False

        if game_size1 == 0xFF and game_size3 == 0xFF and blank_is_valid:
            # This is a blank game settings section.
            return True

        game_section1 = data[44:(44 + game_size1)]
        game_section2 = data[(44 + game_size1):(44 + game_size1 + game_size3)]

        if data[36:38] != NaomiEEPRom.crc(game_section1):
            # The CRC doesn't match!
            return False
        if data[40:42] != NaomiEEPRom.crc(game_section2):
            # The CRC doesn't match!
            return False

        # Everything looks good!
        return True

    def __fix_crc(self) -> None:
        # First fix the system CRCs.
        sys_section1 = self._data[2:18]
        sys_section2 = self._data[20:36]

        if isinstance(self._data, bytes):
            self._data = NaomiEEPRom.crc(sys_section1) + self._data[2:18] + NaomiEEPRom.crc(sys_section2) + self._data[20:]
        elif isinstance(self._data, FileBytes):
            self._data[0:2] = NaomiEEPRom.crc(sys_section1)
            self._data[18:20] = NaomiEEPRom.crc(sys_section2)
        else:
            raise Exception("Logic error!")

        # Now, fix game CRCs.
        game_size1, game_size2 = struct.unpack("<BB", self._data[38:40])
        if game_size1 != game_size2:
            # These numbers should always match. Skip fixing the CRCs if they don't.
            return
        game_size3, game_size4 = struct.unpack("<BB", self._data[42:44])
        if game_size3 != game_size4:
            # These numbers should always match. Skip fixing the CRCs if they don't.
            return

        if game_size1 != 0xFF:
            game_section1 = self._data[44:(44 + game_size1)]

            if isinstance(self._data, bytes):
                self._data = self._data[:36] + NaomiEEPRom.crc(game_section1) + self._data[38:]
            elif isinstance(self._data, FileBytes):
                self._data[36:38] = NaomiEEPRom.crc(game_section1)
            else:
                raise Exception("Logic error!")

            if game_size3 != 0xFF:
                game_section2 = self._data[(44 + game_size1):(44 + game_size1 + game_size3)]
                if isinstance(self._data, bytes):
                    self._data = self._data[:40] + NaomiEEPRom.crc(game_section2) + self._data[42:]
                elif isinstance(self._data, FileBytes):
                    self._data[40:42] = NaomiEEPRom.crc(game_section2)
                else:
                    raise Exception("Logic error!")
            else:
                if isinstance(self._data, bytes):
                    self._data = self._data[:40] + struct.pack("<BB", 0xFF, 0xFF) + self._data[42:]
                elif isinstance(self._data, FileBytes):
                    self._data[40:42] = struct.pack("<BB", 0xFF, 0xFF)
                else:
                    raise Exception("Logic error!")
        else:
            if isinstance(self._data, bytes):
                self._data = self._data[:36] + struct.pack("<BB", 0xFF, 0xFF) + self._data[38:]
                self._data = self._data[:40] + struct.pack("<BB", 0xFF, 0xFF) + self._data[42:]
            elif isinstance(self._data, FileBytes):
                self._data[40:42] = struct.pack("<BB", 0xFF, 0xFF)
                self._data[36:38] = struct.pack("<BB", 0xFF, 0xFF)
            else:
                raise Exception("Logic error!")

        if len(self._data) != 128:
            raise Exception("Logic error!")

    @property
    def data(self) -> BytesLike:
        self.__fix_crc()
        return cast(BytesLike, self._data)

    @property
    def serial(self) -> bytes:
        # Arbitrarily choose the first enclave.
        return self._data[3:7]

    @property
    def system(self) -> ArrayBridge:
        return ArrayBridge(self, NaomiEEPRom.__validate_system, "system", 16, 2, 20)

    @system.setter
    def system(self, data: bytes) -> None:
        bridge = ArrayBridge(self, NaomiEEPRom.__validate_system, "system", 16, 2, 20)
        bridge.data = data

    @property
    def length(self) -> int:
        # Arbitrarily choose the first enclave.
        if not NaomiEEPRom.__validate_game(self._data):
            return 0
        return cast(int, struct.unpack("<B", self._data[38:39])[0])

    @length.setter
    def length(self, newval: int) -> None:
        if newval < 0 or newval > 42:
            raise NaomiEEPRomException("Game section length invalid!")
        if newval == 0:
            # Special case for empty game section.
            newval = 0xFF
        lengthbytes = struct.pack("<BB", newval, newval)
        if isinstance(self._data, bytes):
            self._data = self._data[:38] + lengthbytes + self._data[40:42] + lengthbytes + self._data[44:]
        elif isinstance(self._data, FileBytes):
            self._data[38:40] = lengthbytes
            self._data[42:44] = lengthbytes
        else:
            raise Exception("Logic error!")

        if len(self._data) != 128:
            raise Exception("Logic error!")

        # Force the checksum to be updated so that the "valid" attribute on any
        # ArrayBridge becomes true.
        self.__fix_crc()

    @property
    def game(self) -> ArrayBridge:
        length = self.length
        return ArrayBridge(self, NaomiEEPRom.__validate_game, "game", length, 44, 44 + length)

    @game.setter
    def game(self, data: bytes) -> None:
        length = self.length
        bridge = ArrayBridge(self, NaomiEEPRom.__validate_game, "game", length, 44, 44 + length)
        bridge.data = data

    @overload
    def __getitem__(self, key: int) -> int:
        ...

    @overload
    def __getitem__(self, key: slice) -> bytes:
        ...

    def __getitem__(self, key: Union[int, slice]) -> Union[int, bytes]:
        if isinstance(key, int):
            if key < 0 or key >= 128:
                raise NaomiEEPRomException("Cannot get bytes outside of 128-byte EEPRom!")

            # Arbitrarily choose the first section.
            return self._data[key]

        elif isinstance(key, slice):
            # Determine start of slice
            if key.start is None:
                start = 0
            elif key.start < 0:
                raise Exception("Do not support negative indexing!")
            else:
                start = key.start

            if key.stop is None:
                stop = 128
            elif key.stop < 0:
                raise Exception("Do not support negative indexing!")
            elif key.stop > 128:
                raise NaomiEEPRomException("Cannot get bytes outside of 128-byte EEPRom!")
            else:
                stop = key.stop

            if key.step not in {None, 1}:
                raise NaomiEEPRomException("Cannot get bytes with stride != 1 for 128-byte EEPROM!")

            return self._data[start:stop]

        else:
            raise NotImplementedError("Not implemented!")

    @overload
    def __setitem__(self, key: int, val: int) -> None:
        ...

    @overload
    def __setitem__(self, key: slice, val: bytes) -> None:
        ...

    def __setitem__(self, key: Union[int, slice], val: Union[int, bytes]) -> None:
        if isinstance(key, int):
            if key in {0, 1, 18, 19, 36, 37, 40, 41}:
                raise NaomiEEPRomException("Cannot manually set CRC bytes!")
            if key < 0 or key >= 128:
                raise NaomiEEPRomException("Cannot set bytes outside of 128-byte EEPRom!")
            if isinstance(val, int):
                val = bytes([val])

            if isinstance(self._data, bytes):
                self._data = self._data[:key] + val + self._data[key + 1:]
            elif isinstance(self._data, FileBytes):
                self._data[key:(key + 1)] = val
            else:
                raise Exception("Logic error!")

            if len(self._data) != 128:
                raise Exception("Logic error!")

        else:
            # Determine start of slice
            if key.start is None:
                start = 0
            elif key.start < 0:
                raise Exception("Do not support negative indexing!")
            else:
                start = key.start

            if key.stop is None:
                stop = 128
            elif key.stop < 0:
                raise Exception("Do not support negative indexing!")
            elif key.stop > 128:
                raise NaomiEEPRomException("Cannot set bytes outside of 128-byte EEPRom!")
            else:
                stop = key.stop

            if key.step not in {None, 1}:
                raise NaomiEEPRomException("Cannot set bytes with stride != 1 for 128-byte EEPRom!")

            # Check input to function.
            if not isinstance(val, bytes):
                raise NaomiEEPRomException("Cannot change size of 128-byte EEPRom assigning!")
            if len(val) != (stop - start):
                raise NaomiEEPRomException("Cannot change size of 128-byte EEPRom when assigning!")
            for off in range(start, stop):
                if off in {0, 1, 18, 19, 36, 37, 40, 41}:
                    raise NaomiEEPRomException("Cannot manually set CRC bytes!")

            # Now, set the bytes.
            if isinstance(self._data, bytes):
                self._data = self._data[:start] + val + self._data[stop:]
            elif isinstance(self._data, FileBytes):
                self._data[start:stop] = val
            else:
                raise Exception("Logic error!")

            if len(self._data) != 128:
                raise Exception("Logic error!")



import json
import os
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from arcadeutils import FileBytes
from naomi.eeprom import NaomiEEPRom
from naomi.rom import NaomiRom, NaomiRomRegionEnum
from settings import Settings, SettingsConfig, JSONParseException


class NaomiSettingTypeEnum(Enum):
    UNKNOWN = "unknown"
    SYSTEM = "system"
    GAME = "game"


class NaomiSettingsWrapper:
    # A wrapper class to hold both a system and game settings section together.

    def __init__(self, serial: bytes, system: Settings, game: Settings) -> None:
        self.serial = serial
        self.system = system
        self.game = game

        self.system.type = NaomiSettingTypeEnum.SYSTEM.value
        self.game.type = NaomiSettingTypeEnum.GAME.value

    @staticmethod
    def from_json(settings_files: Dict[str, str], jsondict: Dict[str, Any], context: List[str]) -> "NaomiSettingsWrapper":
        # First, verify that we at least know about the system settings file.
        if "system.settings" not in settings_files:
            raise FileNotFoundError("system.settings does not seem to exist in settings definition directory!")

        # Now, grab the sections that we know we can parse.
        serial = jsondict.get('serial')
        if not isinstance(serial, str) or len(serial) != 4:
            raise JSONParseException(f"\"serial\" key in JSON has invalid data \"{serial}\"!", context)
        serialbytes = serial.encode('ascii')
        gamejson = jsondict.get('game')
        if not isinstance(gamejson, dict):
            raise JSONParseException(f"\"game\" key in JSON has invalid data \"{gamejson}\"!", context)
        systemjson = jsondict.get('system')
        if not isinstance(systemjson, dict):
            raise JSONParseException(f"\"system\" key in JSON has invalid data \"{systemjson}\"!", context)

        # First, load the system settings.
        with open(os.path.join(settings_files["system.settings"]), "r") as fp:
            systemdata = fp.read()
        systemconfig = SettingsConfig.from_data("system.settings", systemdata)

        # Now load the game settings, or if it doesn't exist, default to only allowing system settings to be set.
        gameconfig = NaomiSettingsManager._serial_to_config(settings_files, serialbytes) or SettingsConfig.blank()

        # Finally parse the EEPRom based on the config.
        system = Settings.from_json(NaomiSettingTypeEnum.SYSTEM.value, systemconfig, systemjson, [*context, "system"])
        game = Settings.from_json(NaomiSettingTypeEnum.GAME.value, gameconfig, gamejson, [*context, "game"])
        return NaomiSettingsWrapper(serialbytes, system, game)

    def to_json(self) -> Dict[str, Any]:
        return {
            'serial': self.serial.decode('ascii'),
            'system': self.system.to_json(),
            'game': self.game.to_json(),
        }

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2)

    def __repr__(self) -> str:
        return str(self)


def get_default_settings_directory() -> str:
    # Specifically for projects including this code as a 3rd-party dependency,
    # look up where we stick the default settings definitions files and return
    # that path as a string, suitable for passing into the "directory" param of
    # SettingsManager.
    return os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "definitions"))


class NaomiSettingsManager:
    # A manager class that can handle manifesting and saving settings given a directory
    # of definition files.

    def __init__(self, directory: str) -> None:
        self.__directory = directory

    @property
    def files(self) -> Dict[str, str]:
        return {
            f: os.path.join(self.__directory, f)
            for f in os.listdir(self.__directory)
            if os.path.isfile(os.path.join(self.__directory, f)) and f.endswith(".settings")
        }

    def files_for_serial(self, serial: bytes) -> Dict[str, str]:
        fnames = {f"{serial.decode('ascii')}.settings", "system.settings"}
        return {f: d for (f, d) in self.files.items() if f in fnames}

    def files_for_rom(self, rom: NaomiRom) -> Dict[str, str]:
        return self.files_for_serial(rom.serial)

    def files_for_eeprom(self, data: Union[bytes, FileBytes]) -> Dict[str, str]:
        eeprom = NaomiEEPRom(data)
        return self.files_for_serial(eeprom.serial)

    @staticmethod
    def _serial_to_config(files: Dict[str, str], serial: bytes) -> Optional[SettingsConfig]:
        fname = f"{serial.decode('ascii')}.settings"

        if fname not in files:
            return None

        with open(files[fname], "r") as fp:
            data = fp.read()

        return SettingsConfig.from_data(fname, data)

    def from_serial(self, serial: bytes) -> NaomiSettingsWrapper:
        config = self._serial_to_config(self.files, serial)
        defaults = None
        if config is not None:
            defaults = config.defaults()

        return self.from_eeprom(NaomiEEPRom.default(serial, game_defaults=defaults).data)

    def from_rom(self, rom: NaomiRom, region: NaomiRomRegionEnum) -> NaomiSettingsWrapper:
        # Grab system defaults from ROM header.
        serial = rom.serial
        system_defaults = rom.defaults[region]

        # Grab game defaults from settings file.
        config = self._serial_to_config(self.files, serial)
        game_defaults = None
        if config is not None:
            game_defaults = config.defaults()

        # Create a default EEPROM based on both of the above.
        return self.from_eeprom(NaomiEEPRom.default(serial, system_defaults=system_defaults, game_defaults=game_defaults).data)

    def from_eeprom(self, data: Union[bytes, FileBytes]) -> NaomiSettingsWrapper:
        # First grab the parsed EEPRom so we can get the serial.
        eeprom = NaomiEEPRom(data)

        # Now load the system settings.
        with open(os.path.join(self.__directory, "system.settings"), "r") as fp:
            systemdata = fp.read()
        systemconfig = SettingsConfig.from_data("system.settings", systemdata)

        # Now load the game settings, or if it doesn't exist, default to only
        # allowing system settings to be set.
        gameconfig = self._serial_to_config(self.files, eeprom.serial) or SettingsConfig.blank()

        # Finally parse the EEPRom based on the config.
        system = Settings.from_config(NaomiSettingTypeEnum.SYSTEM.value, systemconfig, eeprom.system.data)
        game = Settings.from_config(NaomiSettingTypeEnum.GAME.value, gameconfig, eeprom.game.data)
        return NaomiSettingsWrapper(eeprom.serial, system, game)

    def from_json(self, jsondict: Dict[str, Any], context: Optional[List[str]] = None) -> NaomiSettingsWrapper:
        return NaomiSettingsWrapper.from_json(self.files, jsondict, context or [])

    def to_eeprom(self, settings: NaomiSettingsWrapper) -> bytes:
        # First, create the EEPROM.
        eeprom = NaomiEEPRom.default(settings.serial)

        # Now, update the game length.
        eeprom.length = settings.game.length

        for section, settingsgroup in [
            (eeprom.system, settings.system),
            (eeprom.game, settings.game),
        ]:
            if not section.valid:
                # If we couldn't make this section correct, completely skip out on it.
                continue

            section.data = settingsgroup.to_bytes()

        return eeprom.data



from settings.settings import (
    SettingsParseException,
    SettingsSaveException,
    JSONParseException,
    ReadOnlyCondition,
    DefaultCondition,
    DefaultConditionGroup,
    Setting,
    Settings,
    SettingsConfig,
)


__all__ = [
    "SettingsParseException",
    "SettingsSaveException",
    "JSONParseException",
    "ReadOnlyCondition",
    "DefaultCondition",
    "DefaultConditionGroup",
    "Setting",
    "Settings",
    "SettingsConfig",
]



#! /usr/bin/env python3
import copy
import os
from dragoncurses.component import (
    Component,
    DeferredInput,
    LabelComponent,
    DialogBoxComponent,
    StickyComponent,
    ClickableComponent,
    SelectInputComponent,
    ButtonComponent,
    BorderComponent,
    ListComponent,
)
from dragoncurses.context import RenderContext, BoundingRectangle
from dragoncurses.scene import Scene
from dragoncurses.loop import execute
from dragoncurses.input import (
    InputEvent,
    KeyboardInputEvent,
    MouseInputEvent,
    Buttons,
    Keys,
)
from dragoncurses.settings import Settings as DragonCursesSettings
from typing import Any, Callable, Dict, List, Tuple, Union

from naomi.settings.settings import NaomiSettingsWrapper
from settings import Settings, ReadOnlyCondition


class ClickableSelectInputComponent(ClickableComponent, SelectInputComponent):
    def __repr__(self) -> str:
        return "ClickableSelectInputComponent(selected={}, options={}, focused={})".format(repr(self.selected), repr(self.options), "True" if self.focus else "False")


class TabComponent(Component):

    def __init__(self, tabs: List[Tuple[str, Component]]) -> None:
        super().__init__()
        self.__buttons = [
            ButtonComponent(
                name,
                formatted=True,
                centered=True,
            ).on_click(lambda component, button: self.__change_tab(component))
            for name, _ in tabs
        ]
        self.__borders = [
            BorderComponent(
                component,
                style=(
                    BorderComponent.SINGLE if DragonCursesSettings.enable_unicode
                    else BorderComponent.ASCII
                ),
            )
            for _, component in tabs
        ]
        self.__tabs = tabs
        self.__selected = 0
        self.__drawn = False
        self.__highlight()

    def __highlight(self) -> None:
        for i, button in enumerate(self.__buttons):
            button.invert = i == self.__selected

    def __change_tab(self, component: Component) -> bool:
        for i, btn in enumerate(self.__buttons):
            if btn is component:
                self.__selected = i
                self.__drawn = False
                self.__highlight()

                return True
        return False

    @property
    def dirty(self) -> bool:
        if not self.__drawn:
            return True
        for component in [*self.__buttons, *self.__borders]:
            if component.dirty:
                return True
        return False

    def attach(self, scene: "Scene", settings: Dict[str, Any]) -> None:
        for component in [*self.__buttons, *self.__borders]:
            component._attach(scene, settings)

    def detach(self) -> None:
        for component in [*self.__buttons, *self.__borders]:
            component._detach()

    def tick(self) -> None:
        for component in [*self.__buttons, *self.__borders]:
            component.tick()

    def handle_input(self, event: "InputEvent") -> Union[bool, DeferredInput]:
        if isinstance(event, KeyboardInputEvent):
            if event.character == Keys.TAB:
                self.__selected = (self.__selected + 1) % len(self.__buttons)
                self.__drawn = False
                self.__highlight()

                return True

        for component in [*self.__buttons, self.__borders[self.__selected]]:
            if component._handle_input(event):
                return True
        return False

    def render(self, context: RenderContext) -> None:
        # Bookkeeping please!
        self.__drawn = True
        context.clear()

        # First, draw the tab buttons.
        for i, button in enumerate(self.__buttons):
            button._render(
                context,
                BoundingRectangle(
                    top=context.bounds.top,
                    bottom=context.bounds.top + 3,
                    left=context.bounds.left + (22 * i),
                    right=context.bounds.left + (22 * i) + 21,
                ),
            )

        # Now, draw the actual component that is selected.
        self.__borders[self.__selected]._render(
            context,
            BoundingRectangle(
                top=context.bounds.top + 3,
                bottom=context.bounds.bottom,
                left=context.bounds.left,
                right=context.bounds.right,
            ),
        )


class SettingsComponent(Component):
    def __init__(self, serial: str, settings: Settings) -> None:
        super().__init__()
        self.__all_settings = settings.settings
        self.__settings = [s for s in settings.settings if s.read_only is not True]
        self.__container: Component
        if settings.settings:
            if self.__settings:
                self.__labels = [
                    LabelComponent(setting.name)
                    for setting in self.__settings
                ]
                self.__inputs = [
                    ClickableSelectInputComponent(
                        setting.values[setting.current or list(setting.values.keys())[0]],
                        [v for _, v in setting.values.items()],
                        focused=False
                    ).on_click(self.__click_select)
                    for setting in self.__settings
                ]
                self.__inputs[0].focus = True
                self.__calculate_visible()
                self.__container = ListComponent(
                    [
                        ListComponent(
                            list(pair),
                            direction=ListComponent.DIRECTION_LEFT_TO_RIGHT,
                        )
                        for pair in zip(self.__labels, self.__inputs)
                    ],
                    direction=ListComponent.DIRECTION_TOP_TO_BOTTOM,
                    size=2,
                )
            else:
                self.__container = LabelComponent(
                    f"Settings definition file \"{serial}.settings\" specifies no editable settings.\n"
                    "As a result, we cannot display or edit game settings for this game!",
                    formatted=True,
                )
        else:
            self.__container = LabelComponent(
                f"Settings definition file \"{serial}.settings\" is missing.\n"
                "As a result, we cannot display or edit game settings for this game!",
                formatted=True,
            )

    def __calculate_visible(self) -> None:
        for i, setting in enumerate(self.__settings):
            if isinstance(setting.read_only, ReadOnlyCondition):
                read_only = setting.read_only.evaluate(self.__all_settings)

                self.__labels[i].visible = not read_only
                self.__inputs[i].visible = not read_only
                if read_only and self.__inputs[i].focus:
                    self.__inputs[i].focus = False
                    if i + 1 < len(self.__inputs):
                        self.__inputs[i + 1].focus = True

    def __click_select(self, component: Component, button: str) -> bool:
        if button == Buttons.LEFT:
            if isinstance(component, ClickableSelectInputComponent):
                if not component.visible:
                    return False

            for inp in self.__inputs:
                inp.focus = inp is component
        # Allow this input to continue propagating, so we can focus on and also click
        # the select option dialog.
        return False

    @property
    def dirty(self) -> bool:
        return self.__container.dirty

    def attach(self, scene: "Scene", settings: Dict[str, Any]) -> None:
        self.__container._attach(scene, settings)

    def detach(self) -> None:
        self.__container._detach()

    def render(self, context: RenderContext) -> None:
        self.__container._render(context, context.bounds)

    def tick(self) -> None:
        self.__container.tick()

    def handle_input(self, event: "InputEvent") -> Union[bool, DeferredInput]:
        if isinstance(event, KeyboardInputEvent):
            if event.character == Keys.UP:
                for i, component in enumerate(self.__inputs):
                    if i > 0 and component.focus:
                        # Find previous setting that isn't invisible.
                        for j in range(i - 1, -1, -1):
                            if self.__inputs[j].visible:
                                component.focus = False
                                self.__inputs[j].focus = True
                                return True
                return True
            if event.character == Keys.DOWN:
                for i, component in enumerate(self.__inputs):
                    if i != (len(self.__inputs) - 1) and component.focus:
                        # Find next setting that isn't invisible.
                        for j in range(i + 1, len(self.__inputs)):
                            if self.__inputs[j].visible:
                                component.focus = False
                                self.__inputs[j].focus = True
                                return True
                return True

        inputhandled = self.__container._handle_input(event)

        for i, setting in enumerate(self.__settings):
            for k, v in setting.values.items():
                if v == self.__inputs[i].selected:
                    setting.current = k
                    break
            else:
                raise Exception(f"Logic error! {self.__inputs[i].selected} {setting.values}")
        self.__calculate_visible()

        return inputhandled


class ClickableLabelComponent(LabelComponent):

    callback = None

    def on_click(self: "ClickableLabelComponent", callback: Callable[["ClickableLabelComponent", MouseInputEvent], bool]) -> "ClickableLabelComponent":
        self.callback = callback
        return self

    def handle_input(self, event: "InputEvent") -> Union[bool, DeferredInput]:
        # Overrides handle_input instead of _handle_input because this is
        # meant to be used as either a mixin. This handles input entirely,
        # instead of intercepting it, so thus overriding the public function.
        if isinstance(event, MouseInputEvent):
            if self.callback is not None:
                handled = self.callback(self, event)
                # Fall through to default if the callback didn't handle.
                if bool(handled):
                    return handled
            else:
                # We still handled this regardless of notification
                return True

        return super().handle_input(event)


class EditorScene(Scene):

    def create(self) -> Component:
        return StickyComponent(
            ClickableLabelComponent(
                "<invert> tab - switch settings tab </invert> "
                "<invert> up/down - select setting </invert> "
                "<invert> left/right - change setting </invert> "
                "<invert> q - quit </invert>",
                formatted=True,
            ).on_click(self.__handle_label_click),
            TabComponent(
                [
                    (
                        "&System Settings",
                        SettingsComponent(self.settings["settings"].serial.decode('ascii'), self.settings["settings"].system),
                    ),
                    (
                        "&Game Settings",
                        SettingsComponent(self.settings["settings"].serial.decode('ascii'), self.settings["settings"].game),
                    ),
                ]
            ),
            location=StickyComponent.LOCATION_BOTTOM,
            size=1,
        )

    def save_eeprom(self) -> None:
        # Copy the edited settings over so that the main loop knows we should
        # save them back.
        self.settings['newsettings'] = copy.deepcopy(self.settings['settings'])
        self.main_loop.exit()

    def __handle_label_click(self, component: Component, event: MouseInputEvent) -> bool:
        # This is hacky, we really should define a component and just handle click
        # events without checking position. However, I don't want to work on a
        # list component that sizes each entry to the entry's width/height.
        if event.button == Buttons.LEFT:
            location = component.location
            if location is not None:
                click_x = event.x - location.left
                click_y = event.y - location.top
                if click_y == 0 and click_x >= 85 and click_x <= 94:
                    self.__display_confirm_quit()
                    return True

        return False

    def __display_confirm_quit(self) -> None:
        self.register_component(
            DialogBoxComponent(
                'Write back changes to EEPROM file?',
                [
                    (
                        '&Yes',
                        lambda c, o: self.save_eeprom(),
                    ),
                    (
                        '&No',
                        lambda c, o: self.main_loop.exit(),
                    ),
                    (
                        '&Cancel',
                        lambda c, o: self.unregister_component(c),
                    ),
                ],
            )
        )

    def handle_input(self, event: InputEvent) -> bool:
        if isinstance(event, KeyboardInputEvent):
            if event.character == 'q':
                self.__display_confirm_quit()
                return True

        return False


class NaomiSettingsEditor:
    def __init__(self, settings: NaomiSettingsWrapper, enable_unicode: bool = True) -> None:
        self.settings = settings
        DragonCursesSettings.enable_unicode = enable_unicode

    def run(self) -> bool:
        # Run the editor, return True if the user requested to save the EEPROM,
        # return False if the user requested to abandon changes.
        os.environ.setdefault('ESCDELAY', '0')

        context = {'settings': copy.deepcopy(self.settings)}
        execute(EditorScene, context)

        # Copy out the values since we want to preserve the input settings by reference.
        if 'newsettings' in context:
            newsettings = context['newsettings']
            for newsetting in newsettings.system.settings:
                for cursetting in self.settings.system.settings:
                    if cursetting.name == newsetting.name:
                        cursetting.current = newsetting.current
                        break
            for newsetting in newsettings.game.settings:
                for cursetting in self.settings.game.settings:
                    if cursetting.name == newsetting.name:
                        cursetting.current = newsetting.current
                        break
            return True
        else:
            return False



import base64
import json
import struct
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union


class SettingSizeEnum(Enum):
    UNKNOWN = auto()
    NIBBLE = auto()
    BYTE = auto()


NO_FILE: str = "NO FILE"


class SettingsParseException(Exception):
    def __init__(self, msg: str, filename: str) -> None:
        super().__init__(msg)
        self.filename = filename


class SettingsSaveException(Exception):
    def __init__(self, msg: str, filename: str) -> None:
        super().__init__(msg)
        self.filename = filename


class JSONParseException(Exception):
    def __init__(self, msg: str, context: List[str]) -> None:
        super().__init__(msg)
        self.context = context


class ReadOnlyCondition:
    # A wrapper class to encapsulate that a setting is read-only based on the
    # value of another setting.

    def __init__(self, filename: str, setting: str, name: str, values: List[int], negate: bool) -> None:
        self.filename = filename
        self.setting = setting
        self.name = name
        self.values = values
        self.negate = negate

    def evaluate(self, settings: List["Setting"]) -> bool:
        for setting in settings:
            if setting.name.lower() == self.name.lower():
                if setting.current in self.values:
                    return self.negate
                else:
                    return not self.negate

        raise SettingsSaveException(
            f"The setting \"{self.setting}\" depends on the value for \"{self.name}\" but that setting does not seem to exist! Perhaps you misspelled \"{self.name}\"?",
            self.filename,
        )


class DefaultCondition:
    # A wrapper class to encapsulate one rule for a setting default.

    def __init__(self, name: str, values: List[int], negate: bool, default: int) -> None:
        self.name = name
        self.values = values
        self.negate = negate
        self.default = default


class DefaultConditionGroup:
    # A wrapper class to encapsulate a set of rules for defaulting a setting.

    def __init__(self, filename: str, setting: str, conditions: List[DefaultCondition]) -> None:
        self.filename = filename
        self.setting = setting
        self.conditions = conditions

    def evaluate(self, filename: str, name: str, settings: List["Setting"]) -> int:
        for cond in self.conditions:
            for setting in settings:
                if setting.name.lower() == cond.name.lower():
                    current = setting.current if setting.current is not None else setting.default

                    if cond.negate and current not in cond.values:
                        return cond.default
                    if not cond.negate and current in cond.values:
                        return cond.default

        namelist = list({f'"{c.name}"' for c in self.conditions})
        if len(namelist) > 2:
            namelist = [", ".join(namelist[:-1]), namelist[-1]]
        names = " or ".join(namelist)

        raise SettingsSaveException(
            f"The default for setting \"{self.setting}\" could not be determined! Perhaps you misspelled one of {names}, or you forgot a value?",
            self.filename,
        )


class Setting:
    # A single setting, complete with its name, size (and optional length if
    # the size is a byte), whether it is read-only, the allowed values for
    # the setting and finally the current value (if it has been parsed out
    # of a valid EEPROM file).

    def __init__(
        self,
        name: str,
        order: int,
        size: SettingSizeEnum,
        length: int,
        read_only: Union[bool, ReadOnlyCondition],
        values: Optional[Dict[int, str]] = None,
        current: Optional[int] = None,
        default: Optional[Union[int, DefaultConditionGroup]] = None,
    ) -> None:
        self.name = name
        self.order = order
        self.size = size
        self.length = length
        self.read_only = read_only
        self.values = values or {}
        self.current = current
        self.default = default

        if size == SettingSizeEnum.UNKNOWN:
            raise Exception("Logic error!")
        if length > 1 and size != SettingSizeEnum.BYTE:
            raise Exception("Logic error!")
        if order < 0:
            raise Exception("Logic error!")

    @staticmethod
    def from_json(filename: str, jsondict: Dict[str, Any], context: List[str]) -> "Setting":
        # Convert JSON dictionary back to a Setting class.
        name = jsondict.get('name')
        if not isinstance(name, str):
            raise JSONParseException(f"\"name\" key in JSON has invalid data \"{name}\"!", context)

        sizestr = jsondict.get('size')
        if not isinstance(sizestr, str):
            raise JSONParseException(f"\"size\" key in JSON has invalid data \"{sizestr}\"!", context)
        try:
            size = SettingSizeEnum[sizestr]
        except KeyError:
            size = SettingSizeEnum.UNKNOWN
        if size == SettingSizeEnum.UNKNOWN:
            raise JSONParseException(f"\"size\" key in JSON has invalid data \"{sizestr}\"!", context)

        length = jsondict.get('length')
        if not isinstance(length, int):
            # Try converting from string.
            try:
                length = int(length)  # type: ignore
            except ValueError:
                pass

            if not isinstance(length, int):
                raise JSONParseException(f"\"length\" key in JSON has invalid data \"{length}\"!", context)

        order = jsondict.get('order')
        if not isinstance(order, int):
            # Try converting from string.
            try:
                order = int(order)  # type: ignore
            except ValueError:
                pass

            if not isinstance(order, int):
                raise JSONParseException(f"\"order\" key in JSON has invalid data \"{order}\"!", context)

        current = jsondict.get('current')
        if current is not None and not isinstance(current, int):
            # Try converting from string.
            try:
                current = int(current)
            except ValueError:
                pass

            if not isinstance(current, int):
                raise JSONParseException(f"\"current\" key in JSON has invalid data \"{current}\"!", context)

        valuedict = jsondict.get('values')
        values: Optional[Dict[int, str]] = None
        if valuedict is not None:
            if not isinstance(valuedict, dict):
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{valuedict}\"!", context)
            try:
                values = {int(k): str(v) for (k, v) in valuedict.items()}
            except ValueError:
                pass
            if values is None:
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{valuedict}\"!", context)

        readonlydict = jsondict.get('readonly')
        readonly: Union[bool, ReadOnlyCondition]
        if readonlydict is True:
            readonly = True
        elif readonlydict is False:
            readonly = False
        elif isinstance(readonlydict, dict):
            # Parse out conditional.
            roname = readonlydict.get('name')
            if not isinstance(roname, str):
                raise JSONParseException(f"\"name\" key in JSON has invalid data \"{roname}\"!", [*context, "readonly"])

            rovaluelist = readonlydict.get('values')
            if not isinstance(rovaluelist, list):
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{rovaluelist}\"!", [*context, "readonly"])
            rovalues = None
            try:
                rovalues = [int(k) for k in rovaluelist]
            except ValueError:
                pass
            if rovalues is None:
                raise JSONParseException(f"\"values\" key in JSON has invalid data \"{rovaluelist}\"!", [*context, "readonly"])

            ronegate = readonlydict.get('negate')
            if not isinstance(ronegate, bool):
                raise JSONParseException(f"\"negate\" key in JSON has invalid data \"{ronegate}\"!", [*context, "readonly"])

            readonly = ReadOnlyCondition(filename, name, roname, rovalues, ronegate)
        else:
            raise JSONParseException(f"\"readonly\" key in JSON has invalid data \"{readonlydict}\"!", context)

        defaultdict = jsondict.get('default')
        default: Optional[Union[int, DefaultConditionGroup]] = None
        if defaultdict is None or isinstance(defaultdict, int):
            default = defaultdict
        elif isinstance(defaultdict, str):
            try:
                default = int(defaultdict)
            except ValueError:
                default = None

            if default is None:
                raise JSONParseException(f"\"default\" key in JSON has invalid data \"{defaultdict}\"!", context)
        elif isinstance(defaultdict, list):
            # Go through each entry in the list and parse out the conditionals.
            default = DefaultConditionGroup(filename, name, [])

            for i, defaultblob in enumerate(defaultdict):
                # Parse out conditional.
                defname = defaultblob.get('name')
                if not isinstance(defname, str):
                    raise JSONParseException(f"\"name\" key in JSON has invalid data \"{defname}\"!", [*context, f"default[{i}]"])

                defvaluelist = defaultblob.get('values')
                if not isinstance(defvaluelist, list):
                    raise JSONParseException(f"\"values\" key in JSON has invalid data \"{defvaluelist}\"!", [*context, f"default[{i}]"])
                defvalues = None
                try:
                    defvalues = [int(k) for k in defvaluelist]
                except ValueError:
                    pass
                if defvalues is None:
                    raise JSONParseException(f"\"values\" key in JSON has invalid data \"{defvaluelist}\"!", [*context, f"default[{i}]"])

                defnegate = defaultblob.get('negate')
                if not isinstance(defnegate, bool):
                    raise JSONParseException(f"\"negate\" key in JSON has invalid data \"{defnegate}\"!", [*context, f"default[{i}]"])

                defdefault = defaultblob.get('default')
                if not isinstance(defdefault, int):
                    # Try converting from string.
                    try:
                        defdefault = int(defdefault)
                    except ValueError:
                        pass

                    if not isinstance(defdefault, int):
                        raise JSONParseException(f"\"default\" key in JSON has invalid data \"{defdefault}\"!", [*context, f"default[{i}]"])

                default.conditions.append(DefaultCondition(defname, defvalues, defnegate, defdefault))
        else:
            raise JSONParseException(f"\"default\" key in JSON has invalid data \"{defaultdict}\"!", context)

        return Setting(
            name,
            order,
            size,
            length,
            readonly,
            values,
            current,
            default,
        )

    def to_json(self) -> Dict[str, Any]:
        jdict = {
            'name': self.name,
            'order': self.order,
            'size': self.size.name,
            'length': self.length,
            'values': self.values,
            'current': self.current,
        }

        if self.read_only is True:
            jdict['readonly'] = True
        elif self.read_only is False:
            jdict['readonly'] = False
        elif isinstance(self.read_only, ReadOnlyCondition):
            jdict['readonly'] = {
                "name": self.read_only.name,
                "values": self.read_only.values,
                "negate": self.read_only.negate,
            }

        if self.default is None or isinstance(self.default, int):
            jdict['default'] = self.default
        elif isinstance(self.default, DefaultConditionGroup):
            jdict['default'] = [
                {
                    "name": cond.name,
                    "values": cond.values,
                    "default": cond.default,
                    "negate": cond.negate,
                }
                for cond in self.default.conditions
            ]
        return jdict

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2)

    def __repr__(self) -> str:
        return str(self)


class Settings:
    # A collection of settings as well as the type of settings this is. This is also responsible
    # for parsing and creating sections in an actual EEPROM file based on the settings themselves.

    def __init__(self, filename: str, settings: List[Setting], type: Optional[str] = None) -> None:
        self.filename = filename
        self.settings = settings
        self.type = type

    @staticmethod
    def from_config(type: str, config: "SettingsConfig", data: bytes) -> "Settings":
        # Keep track of how many bytes into the EEPROM we are.
        location = 0
        halves = 0

        # We need to make sure this is in load order, so that we can parse out the
        # correct settings in order of them showing up in the file.
        for setting in sorted(config.settings, key=lambda setting: setting.order):
            if setting.size == SettingSizeEnum.NIBBLE:
                if location < len(data):
                    if halves == 0:
                        setting.current = (data[location] >> 4) & 0xF
                    else:
                        setting.current = data[location] & 0xF

                if halves == 0:
                    halves = 1
                else:
                    halves = 0
                    location += 1
            elif setting.size == SettingSizeEnum.BYTE:
                if halves != 0:
                    raise SettingsParseException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", config.filename)

                if setting.length == 1:
                    if location < len(data):
                        setting.current = struct.unpack("<B", data[location:(location + 1)])[0]
                elif setting.length == 2:
                    if location < (len(data) - 1):
                        setting.current = struct.unpack("<H", data[location:(location + 2)])[0]
                elif setting.length == 4:
                    if location < (len(data) - 3):
                        setting.current = struct.unpack("<I", data[location:(location + 4)])[0]
                else:
                    raise SettingsParseException(f"Cannot parse setting \"{setting.name}\" with unrecognized size \"{setting.length}\"!", config.filename)

                location += setting.length

        # Return this in the original order presented to us.
        final_settings = Settings(config.filename, config.settings, type=type)
        if final_settings.length != len(data):
            raise SettingsParseException(
                f"Unexpected final size of {type} section, expected {len(data)} bytes but definition file covers {final_settings.length} bytes!",
                config.filename,
            )
        return final_settings

    @property
    def length(self) -> int:
        halves = 0
        length = 0
        for setting in sorted(self.settings, key=lambda setting: setting.order):
            if setting.size == SettingSizeEnum.NIBBLE:
                # Update our length.
                if halves == 0:
                    halves = 1
                else:
                    halves = 0
                    length += 1

            elif setting.size == SettingSizeEnum.BYTE:
                # First, make sure we aren't in a pending nibble state.
                if halves != 0:
                    raise SettingsSaveException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", self.filename)

                if setting.length not in {1, 2, 4}:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with unrecognized size {setting.length}!", self.filename)

                # Update our length.
                length += setting.length

        return length

    @staticmethod
    def from_json(type: str, config: "SettingsConfig", jsondict: Dict[str, Any], context: List[str]) -> "Settings":
        # First, parse out the keys that we know we need.
        typestr = jsondict.get('type')
        if not isinstance(typestr, str) or typestr != type:
            raise JSONParseException(f"\"type\" key in JSON has invalid data \"{typestr}\"!", context)
        filename = jsondict.get('filename')
        if not isinstance(filename, str) or filename != config.filename:
            raise JSONParseException(f"\"filename\" key in JSON has invalid data \"{filename}\"!", context)
        settings = jsondict.get('settings')
        if not isinstance(settings, list):
            raise JSONParseException(f"\"settings\" key in JSON has invalid data \"{settings}\"!", context)

        # Now, parse out the settings in the dict.
        parsedsettings = [Setting.from_json(config.filename, s, [*context, f"settings[{i}]"]) for (i, s) in enumerate(settings)]

        # Finally, go through our config and match up settings to their values.
        for setting in config.settings:
            for parsedsetting in parsedsettings:
                if parsedsetting.name.lower() == setting.name.lower():
                    # It's a match
                    setting.current = parsedsetting.current
                    break
            else:
                raise JSONParseException(f"Setting \"{setting.name}\" could not be found in JSON!", context)

        return Settings(config.filename, config.settings, type=type)

    def to_json(self) -> Dict[str, Any]:
        return {
            'type': self.type,
            'filename': self.filename if self.filename != NO_FILE else None,
            'settings': [
                setting.to_json() for setting in self.settings
            ],
        }

    def to_bytes(self) -> bytes:
        pending: int = 0
        halves: int = 0
        location: int = 0
        section: bytes = b''

        for setting in sorted(self.settings, key=lambda setting: setting.order):
            # First, calculate what the default should be in case we need to use it.
            if setting.default is None:
                default = None
            elif isinstance(setting.default, int):
                default = setting.default
            elif isinstance(setting.default, DefaultConditionGroup):
                # Must evaluate settings to figure out the default for this.
                default = setting.default.evaluate(self.filename, setting.name, self.settings)

            # Now, figure out if we should defer to the default over the current value
            # (if it is read-only) or if we should use the current value.
            if isinstance(setting.read_only, ReadOnlyCondition):
                read_only = setting.read_only.evaluate(self.settings)
            elif setting.read_only is True:
                read_only = True
            elif setting.read_only is False:
                read_only = False

            if read_only:
                # If it is read-only, then only take the current value if the default doesn't
                # exist. This lets settings that are selectively read-only get a conditional
                # default if one exists that takes precedence over the current value.
                value = setting.current if default is None else default
            else:
                # If the setting is not read-only, then only take the default if the current
                # value is None.
                value = default if setting.current is None else setting.current

            # Now, write out the setting by updating the EEPROM in the correct location.
            if setting.size == SettingSizeEnum.NIBBLE:
                # First, if we have anything to write, write it.
                if value is not None:
                    if halves == 0:
                        pending = (value & 0xF) << 4
                    else:
                        if len(section) != location:
                            raise Exception("Logic error!")
                        section += struct.pack("<B", (value & 0xF) | pending)
                else:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with a null value!", self.filename)

                # Now, update our position.
                if halves == 0:
                    halves = 1
                else:
                    halves = 0
                    location += 1

            elif setting.size == SettingSizeEnum.BYTE:
                # First, make sure we aren't in a pending nibble state.
                if halves != 0:
                    raise SettingsSaveException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", self.filename)

                if setting.length not in {1, 2, 4}:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with unrecognized size {setting.length}!", self.filename)

                if value is not None:
                    if len(section) != location:
                        raise Exception("Logic error!")
                    if setting.length == 1:
                        section += struct.pack("<B", value)
                    elif setting.length == 2:
                        section += struct.pack("<H", value)
                    elif setting.length == 4:
                        section += struct.pack("<I", value)
                else:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with a null value!", self.filename)

                # Now, update our position.
                location += setting.length

        return section

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2)

    def __repr__(self) -> str:
        return str(self)


class SettingsConfig:
    # A class that can manifest a list of settings given a particular
    # file. It is not responsible for parsing any settings. It is only
    # responsible for creating the list of settings given a settings
    # definition file.

    def __init__(self, filename: str, settings: List[Setting]) -> None:
        self.filename = filename
        self.settings = settings

    @staticmethod
    def blank() -> "SettingsConfig":
        # It would be weird to display "NO FILE" to the user when there is
        # an error, but virtually all errors arise from parsing the settings
        # file itself, so if this is blank its unlikely errors will happen.
        return SettingsConfig(NO_FILE, [])

    @staticmethod
    def __escaped(data: str) -> str:
        escaped: bool = False
        output: str = ""

        for c in data:
            if escaped:
                output += f"__ESCAPED__{base64.b64encode(c.encode('utf-8')).decode('ascii')}__"
                escaped = False
            elif c == "\\":
                escaped = True
            else:
                output += c
        return output

    @staticmethod
    def __unescaped(data: str) -> str:
        while "__ESCAPED__" in data:
            before, after = data.split("__ESCAPED__", 1)
            if "__" not in after:
                raise Exception("Logic error, expected second __ to find escaped character!")

            escaped, after = after.split("__", 1)
            data = before + base64.b64decode(escaped.encode('ascii')).decode('utf-8') + after
        return data

    @staticmethod
    def __get_kv(filename: str, name: str, setting: str, length: int) -> Dict[int, str]:
        realsetting = setting.strip()
        if realsetting.startswith("values are "):
            realsetting = realsetting[11:]
        if realsetting.startswith("value is "):
            realsetting = realsetting[9:]

        hex_display = False
        if realsetting.endswith(" in hex"):
            realsetting = realsetting[:-7]
            hex_display = True

        def format_val(val: int) -> str:
            nonlocal hex_display
            nonlocal length

            if not hex_display:
                return str(val)

            strval = hex(val)[2:]
            while len(strval) < length:
                strval = "0" + strval
            return strval

        try:
            if "-" in realsetting:
                if " to " in realsetting:
                    raise SettingsParseException(
                        f"Setting \"{name}\" cannot have a range for valid values that includes a dash! \"{setting}\" should be specified like \"20 to E0\".",
                        filename,
                    )

                k, v = realsetting.split("-", 1)
                k = k.strip().replace(" ", "").replace("\t", "")
                key = int(k, 16)
                value = SettingsConfig.__unescaped(v.strip())

                return {key: value}
            else:
                if " to " in realsetting:
                    low, high = realsetting.split(" to ", 1)
                    low = low.strip()
                    high = high.strip()
                    low = low.strip().replace(" ", "").replace("\t", "")
                    high = high.strip().replace(" ", "").replace("\t", "")

                    retdict: Dict[int, str] = {}
                    for x in range(int(low, 16), int(high, 16) + 1):
                        retdict[x] = format_val(x)
                    return retdict
                else:
                    key = int(realsetting.strip(), 16)
                    value = format_val(key)

                    return {key: value}
        except ValueError:
            raise SettingsParseException(
                f"Failed to parse setting \"{name}\", could not understand value \"{setting}\".",
                filename,
            )

    @staticmethod
    def __get_vals(filename: str, name: str, setting: str) -> Tuple[str, List[int]]:
        try:
            name, rest = setting.split(" is ", 1)
            name = SettingsConfig.__unescaped(name.strip())
            vals: List[int] = []

            for val in rest.split(" or "):
                val = val.strip().replace(" ", "").replace("\t", "")
                vals.append(int(val, 16))

            return name, vals
        except ValueError:
            raise SettingsParseException(
                f"Failed to parse setting \"{name}\", could not understand if condition \"{setting}\".",
                filename,
            )

    @staticmethod
    def from_data(filename: str, data: str) -> "SettingsConfig":
        rawlines = data.splitlines()
        lines: List[str] = []
        settings: List[Setting] = []

        for line in rawlines:
            if not line.strip():
                # Ignore empty lines.
                continue
            if line.strip().startswith("#"):
                # Ignore comments.
                continue

            if ":" not in line:
                # Assume that this is a setting entry.
                if not lines:
                    raise SettingsParseException(f"Missing setting name before size, read-only specifier, defaults or value in \"{line}\". Perhaps you forgot a colon?", filename)

                cur = lines[-1]
                if cur.strip()[-1] == ":":
                    cur = cur + line
                else:
                    cur = cur + "," + line

                lines[-1] = cur
            else:
                # Assume that this is a full setting.
                lines.append(line)

        # Keep track of the order settings were loaded, as this is the order they get parsed in.
        order = 0

        pending_insertions: List[Tuple[Tuple[str, str], Setting]] = []

        # Make sure we respect escaping for any character.
        lines = [SettingsConfig.__escaped(line) for line in lines]
        for line in lines:
            # First, get the name as well as the size and any restrictions.
            name, rest = line.split(":", 1)
            name = SettingsConfig.__unescaped(name.strip())
            rest = rest.strip()

            # Now, figure out what size it should be.
            size = SettingSizeEnum.UNKNOWN
            length = 1
            read_only: Union[bool, ReadOnlyCondition] = False
            values: Dict[int, str] = {}
            default: Optional[Union[int, DefaultConditionGroup]] = None
            ordering: Optional[Tuple[str, str]] = None

            if "," in rest:
                restbits = [r.strip() for r in rest.split(",")]
            else:
                restbits = [rest]

            for bit in restbits:
                if "byte" in bit or "nibble" in bit or "half-byte" in bit:
                    if " " in bit:
                        lenstr, units = bit.split(" ", 1)
                        length = int(lenstr.strip())
                        units = units.strip()
                    else:
                        units = bit.strip()

                    if "nibble" in units or "half-byte" in units:
                        size = SettingSizeEnum.NIBBLE
                    elif "byte" in units:
                        size = SettingSizeEnum.BYTE
                    else:
                        raise SettingsParseException(f"Unrecognized unit \"{units}\" for setting \"{name}\". Perhaps you misspelled \"byte\" or \"half-byte\"?", filename)
                    if size != SettingSizeEnum.BYTE and length != 1:
                        raise SettingsParseException(f"Invalid length \"{length}\" for setting \"{name}\". You should only specify a length for bytes.", filename)

                elif "read-only" in bit:
                    condstr = None
                    if " if " in bit:
                        readonlystr, condstr = bit.split(" if ", 1)
                        negate = True
                    elif " unless " in bit:
                        readonlystr, condstr = bit.split(" unless ", 1)
                        negate = False
                    else:
                        # Its unconditionally read-only.
                        read_only = True
                        readonlystr = bit

                    if readonlystr.strip() != "read-only":
                        raise SettingsParseException(f"Cannot parse read-only condition \"{bit}\" for setting \"{name}\"!", filename)
                    if condstr is not None:
                        condname, condvalues = SettingsConfig.__get_vals(filename, name, condstr)
                        read_only = ReadOnlyCondition(filename, name, condname, condvalues, negate)

                elif "default" in bit:
                    if " is " in bit:
                        defstr, rest = bit.split(" is ", 1)
                        if defstr.strip() != "default":
                            raise SettingsParseException(f"Cannot parse default \"{bit}\" for setting \"{name}\"!", filename)

                        condstr = None
                        if " if " in rest:
                            rest, condstr = rest.split(" if ", 1)
                            negate = False
                        elif " unless " in rest:
                            rest, condstr = rest.split(" unless ", 1)
                            negate = True
                        else:
                            # Its unconditionally a default.
                            pass

                        # Now, see if the value is computed from another setting.
                        rest = rest.strip()
                        if "value of " in rest:
                            valname = rest[9:].strip()
                            adjust = 0

                            if "+" in valname:
                                newname, add = valname.rsplit('+', 1)
                                newname = newname.strip()
                                add = add.strip()

                                try:
                                    adjust = int(add, 16)
                                    valname = newname
                                except ValueError:
                                    # Wasn't a correct adjustment, assume its the name.
                                    pass

                            if "-" in valname:
                                newname, add = valname.rsplit('-', 1)
                                newname = newname.strip()
                                add = add.strip()

                                try:
                                    adjust = -int(add, 16)
                                    valname = newname
                                except ValueError:
                                    # Wasn't a correct adjustment, assume its the name.
                                    pass

                            valname = SettingsConfig.__unescaped(valname)

                            if default is None:
                                default = DefaultConditionGroup(filename, name, [])
                            if not isinstance(default, DefaultConditionGroup):
                                raise SettingsParseException(f"Cannot specify an unconditional default alongside conditional defaults for setting \"{name}\"!", filename)

                            def getmax(size: SettingSizeEnum, length: int) -> int:
                                if size == SettingSizeEnum.NIBBLE:
                                    return 0x10
                                elif size == SettingSizeEnum.BYTE:
                                    if length == 1:
                                        return 0x100
                                    elif length == 2:
                                        return 0x10000
                                    elif length == 4:
                                        return 0x100000000

                                raise SettingsParseException(
                                    f"Cannot determine default \"{bit}\" values for setting \"{name}\"!",
                                    filename,
                                )

                            if condstr is None:
                                # This is just a series of "default is <X + adjust> if <valname> is <X>" internally.
                                maxval = getmax(size, length)
                                for x in range(maxval):
                                    default.conditions.append(DefaultCondition(valname, [x], False, (x + adjust) % maxval))
                            else:
                                condname, condvalues = SettingsConfig.__get_vals(filename, name, condstr)

                                if condname != valname:
                                    # This is a crazy complicated predicate where we would need to check if one setting is a certain
                                    # value and if so scoop up the value of another setting. Hopefully we don't need to support this.
                                    raise SettingsParseException(
                                        f"Cannot handle extracting value from \"{valname}\" while being dependent on \"{condname}\" for setting \"{name}\"!",
                                        filename,
                                    )

                                # This is just the same series as above, but with a specific set of X instead of a full range.
                                maxval = getmax(size, length)
                                for x in condvalues:
                                    default.conditions.append(DefaultCondition(valname, [x], False, (x + adjust) % maxval))
                        else:
                            rest = rest.strip().replace(" ", "").replace("\t", "")
                            defbytes = bytes([int(rest[i:(i + 2)], 16) for i in range(0, len(rest), 2)])
                            if size != SettingSizeEnum.UNKNOWN and len(defbytes) == 1:
                                defaultint = defbytes[0]
                            else:
                                if size == SettingSizeEnum.NIBBLE:
                                    defaultint = struct.unpack("<B", defbytes[0:1])[0]
                                elif size == SettingSizeEnum.BYTE:
                                    if length == 1:
                                        defaultint = struct.unpack("<B", defbytes[0:1])[0]
                                    elif length == 2:
                                        defaultint = struct.unpack("<H", defbytes[0:2])[0]
                                    elif length == 4:
                                        defaultint = struct.unpack("<I", defbytes[0:4])[0]
                                    else:
                                        raise SettingsParseException(
                                            f"Cannot convert default \"{bit}\" for setting \"{name}\" because we don't know how to handle length \"{length}\"!",
                                            filename,
                                        )
                                else:
                                    raise SettingsParseException(f"Must place default \"{bit}\" after size specifier in setting \"{name}\"!", filename)

                            if condstr is None:
                                if default is not None:
                                    if isinstance(default, DefaultConditionGroup):
                                        raise SettingsParseException(f"Cannot specify an unconditional default alongside conditional defaults for setting \"{name}\"!", filename)
                                    else:
                                        raise SettingsParseException(f"Cannot specify more than one default for setting \"{name}\"!", filename)
                                default = defaultint
                            else:
                                if default is None:
                                    default = DefaultConditionGroup(filename, name, [])
                                if not isinstance(default, DefaultConditionGroup):
                                    raise SettingsParseException(f"Cannot specify an unconditional default alongside conditional defaults for setting \"{name}\"!", filename)

                                condname, condvalues = SettingsConfig.__get_vals(filename, name, condstr)
                                default.conditions.append(DefaultCondition(condname, condvalues, negate, defaultint))
                    else:
                        raise SettingsParseException(f"Cannot parse default for setting \"{name}\"! Specify defaults like \"default is 0\".", filename)

                elif "display" in bit:
                    # This is a directive for controlling where the setting is displayed.
                    if " before " in bit:
                        display, settingname = bit.split(" before ", 1)
                        display = display.strip()
                        settingname = SettingsConfig.__unescaped(settingname.strip())
                        directive = "before"
                    elif " after " in bit:
                        display, settingname = bit.split(" after ", 1)
                        display = display.strip()
                        settingname = SettingsConfig.__unescaped(settingname.strip())
                        directive = "after"
                    else:
                        raise SettingsParseException(f"Couldn't understand position \"{bit}\" for setting \"{name}\". Perhaps you misspelled \"before\" or \"after\"?", filename)

                    if display != "display":
                        raise SettingsParseException(f"Couldn't understand position \"{bit}\" for setting \"{name}\".", filename)

                    ordering = (directive, settingname)

                else:
                    # Assume this is a setting value.
                    values.update(SettingsConfig.__get_kv(filename, name, bit, 1 if size == SettingSizeEnum.NIBBLE else (length * 2)))

            if size == SettingSizeEnum.UNKNOWN:
                raise SettingsParseException(f"Setting \"{name}\" is missing a size specifier!", filename)
            if read_only is not True and not values:
                raise SettingsParseException(f"Setting \"{name}\" is missing any valid values!", filename)

            # Add it to either the current list of settings, or the queue of settings to insert when we're done.
            new_setting = Setting(
                name,
                order,
                size,
                length,
                read_only,
                values,
                default=default,
            )
            if ordering is None:
                settings.append(new_setting)
            else:
                pending_insertions.append((ordering, new_setting))

            # Keep track of order
            order += 1

        # Verify that nibbles come in pairs.
        halves = 0
        for setting in settings:
            if setting.size == SettingSizeEnum.NIBBLE:
                halves = 1 - halves
            elif setting.size == SettingSizeEnum.BYTE:
                if halves != 0:
                    raise SettingsParseException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", filename)

        # Now, insert pending settings where they belong.
        while pending_insertions:
            # Search all pending insertions to see if we can add one of them.
            # Its okay for pending insertions to depend on the placement of other
            # pending insertions as long as one of them is placeable each time.
            for i in range(len(pending_insertions)):
                (directive, settingname), needs_placement = pending_insertions[i]
                location: Optional[int] = None

                # Look for a potential insertion spot.
                for j, setting in enumerate(settings):
                    if setting.name == settingname:
                        if directive == "before":
                            location = j
                        elif directive == "after":
                            location = j + 1
                        else:
                            raise Exception("Logic error!")
                        break

                if location is not None:
                    # Get rid of the pending setting, we're about to add it to the master list.
                    del pending_insertions[i]
                    settings = [*settings[:location], needs_placement, *settings[location:]]
                    break
            else:
                allsettings = ", ".join([f'"{s.name}"' for _, s in pending_insertions])
                raise SettingsParseException(
                    f"We couldn't figure out where to place the following settings: {allsettings}. Did you accidentially create a display order loop?",
                    filename,
                )

        return SettingsConfig(filename, settings)

    def defaults(self) -> bytes:
        pending = 0
        halves = 0
        defaults: List[bytes] = []

        for setting in sorted(self.settings, key=lambda setting: setting.order):
            if setting.default is None:
                default = 0
            elif isinstance(setting.default, int):
                default = setting.default
            elif isinstance(setting.default, DefaultConditionGroup):
                # Must evaluate settings to figure out the default for this.
                default = setting.default.evaluate(self.filename, setting.name, self.settings)

            if setting.size == SettingSizeEnum.NIBBLE:
                if halves == 0:
                    pending = (default & 0xF) << 4
                else:
                    defaults.append(bytes([(default & 0xF) | pending]))

                if halves == 0:
                    halves = 1
                else:
                    halves = 0
            elif setting.size == SettingSizeEnum.BYTE:
                if halves != 0:
                    raise SettingsSaveException(f"The setting \"{setting.name}\" follows a lonesome half-byte. Half-byte settings must always be in pairs!", self.filename)
                if setting.length == 1:
                    defaults.append(struct.pack("<B", default))
                elif setting.length == 2:
                    defaults.append(struct.pack("<H", default))
                elif setting.length == 4:
                    defaults.append(struct.pack("<I", default))
                else:
                    raise SettingsSaveException(f"Cannot save setting \"{setting.name}\" with unrecognized size {setting.length}!", self.filename)

        return b"".join(defaults)



#! /usr/bin/env python3
import os
import struct
from typing import Generic, Optional, Tuple, TypeVar, Union, cast, overload

from arcadeutils import FileBytes
from naomi import NaomiRom, NaomiRomSection, NaomiEEPRom


def get_default_trojan() -> bytes:
    # Specifically for projects including this code as a 3rd-party dependency,
    # look up where we stick the default already-compiled trojan and return it
    # as bytes that can be passed into the "trojan" param of the
    # NaomiSettingsPatcher constructor. This way, they don't need to know our
    # internal directory layout.
    root = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".."))
    trojan = os.path.join(root, 'homebrew', 'settingstrojan', 'settingstrojan.bin')
    with open(trojan, "rb") as bfp:
        return bfp.read()


@overload
def add_or_update_section(
    data: bytes,
    location: int,
    newsection: bytes,
    header: Optional[NaomiRom] = None,
    verbose: bool = False,
) -> bytes:
    ...


@overload
def add_or_update_section(
    data: FileBytes,
    location: int,
    newsection: bytes,
    header: Optional[NaomiRom] = None,
    verbose: bool = False,
) -> FileBytes:
    ...


def add_or_update_section(
    data: Union[bytes, FileBytes],
    location: int,
    newsection: bytes,
    header: Optional[NaomiRom] = None,
    verbose: bool = False,
) -> Union[bytes, FileBytes]:
    # Note that if an external header is supplied, it will be modified to match the updated
    # data.
    if isinstance(data, FileBytes):
        data = data.clone()
    if header is None:
        header = NaomiRom(data)

    # First, find out if there's already a section in this header.
    executable = header.main_executable
    for section in executable.sections:
        if section.load_address == location:
            # This is a section chunk
            if section.length != len(newsection):
                raise NaomiSettingsPatcherException("Found section in executable, but it is the wrong size!")

            if verbose:
                print("Overwriting old section in existing ROM section.")

            # We can just update the data to overwrite this section
            if isinstance(data, bytes):
                data = data[:section.offset] + newsection + data[(section.offset + section.length):]
            elif isinstance(data, FileBytes):
                data[section.offset:(section.offset + section.length)] = newsection
            else:
                raise Exception("Logic error!")
            break
    else:
        # We need to add an init section to the ROM
        if len(executable.sections) >= 8:
            raise NaomiSettingsPatcherException("ROM already has the maximum number of sections!")

        # Add a new section to the end of the rom for this binary data.
        if verbose:
            print("Attaching section to a new ROM section at the end of the file.")

        # Add a new section to the end of the rom for this section
        executable.sections.append(
            NaomiRomSection(
                offset=len(data),
                load_address=location,
                length=len(newsection)
            )
        )
        header.main_executable = executable

        # Now, just append it to the end of the file
        if isinstance(data, bytes):
            data = header.data + data[header.HEADER_LENGTH:] + newsection
        elif isinstance(data, FileBytes):
            data[:header.HEADER_LENGTH] = header.data
            data.append(newsection)
        else:
            raise Exception("Logic error!")

    # Return the updated data.
    return data


def __is_config(data: Union[bytes, FileBytes], index: int) -> bool:
    if all(x == 0xEE for x in data[index:(index + 4)]) and all(x == 0xEE for x in data[(index + 24):(index + 28)]):
        _, _, _, debug, _ = struct.unpack("<IIIII", data[(index + 4):(index + 24)])
        if debug not in {0, 1, 0xDDDDDDDD}:
            return False
        return True
    return False


def __extract_config(data: Union[bytes, FileBytes], index: int) -> Tuple[int, int, bool, Tuple[int, int, int]]:
    original_start, trojan_start, _, debug, date = struct.unpack("<IIIII", data[(index + 4):(index + 24)])

    day = date % 100
    month = (date // 100) % 100
    year = (date // 10000)

    return (
        original_start,
        trojan_start,
        debug != 0,
        (year, month, day),
    )


def get_config(data: Union[bytes, FileBytes], *, start: Optional[int] = None, end: Optional[int] = None) -> Tuple[int, int, bool, Tuple[int, int, int]]:
    # Returns a tuple consisting of the original EXE start address and
    # the desired trojan start address and whether debug printing is
    # enabled, and the date string of the trojan we're using.
    if isinstance(data, bytes):
        for i in range(start or 0, (end or len(data)) - (28 - 1)):
            if __is_config(data, i):
                return __extract_config(data, i)

    elif isinstance(data, FileBytes):
        start = start or 0
        end = end or len(data)

        while start < end:
            location = data.search(b"\xEE" * 4, start=start, end=end)
            if location is None:
                break

            if __is_config(data, location):
                return __extract_config(data, location)
            start = location + 1

    else:
        raise Exception("Logic error!")

    raise NaomiSettingsPatcherException("Couldn't find config in executable!")


@overload
def change(binfile: bytes, tochange: bytes, loc: int) -> bytes:
    ...


@overload
def change(binfile: FileBytes, tochange: bytes, loc: int) -> FileBytes:
    ...


def change(binfile: Union[bytes, FileBytes], tochange: bytes, loc: int) -> Union[bytes, FileBytes]:
    if isinstance(binfile, bytes):
        return binfile[:loc] + tochange + binfile[(loc + len(tochange)):]
    elif isinstance(binfile, FileBytes):
        binfile[loc:(loc + len(tochange))] = tochange
        return binfile
    else:
        raise Exception("Logic error!")


@overload
def patch_bytesequence(data: bytes, sentinel: int, replacement: bytes) -> bytes:
    ...


@overload
def patch_bytesequence(data: FileBytes, sentinel: int, replacement: bytes) -> FileBytes:
    ...


def patch_bytesequence(data: Union[bytes, FileBytes], sentinel: int, replacement: bytes) -> Union[bytes, FileBytes]:
    if isinstance(data, bytes):
        length = len(replacement)
        for i in range(len(data) - (length - 1)):
            if all(x == sentinel for x in data[i:(i + length)]):
                return change(data, replacement, i)

        raise NaomiSettingsPatcherException("Couldn't find spot to patch in data!")

    elif isinstance(data, FileBytes):
        replace_loc = data.search(bytes(sentinel) * len(replacement))
        if replace_loc is not None:
            return change(data, replacement, replace_loc)

        raise NaomiSettingsPatcherException("Couldn't find spot to patch in data!")

    else:
        raise Exception("Logic error!")


@overload
def add_or_update_trojan(
    data: bytes,
    trojan: bytes,
    debug: int,
    options: int,
    datachunk: Optional[bytes] = None,
    header: Optional[NaomiRom] = None,
    verbose: bool = False,
) -> bytes:
    ...


@overload
def add_or_update_trojan(
    data: FileBytes,
    trojan: bytes,
    debug: int,
    options: int,
    datachunk: Optional[bytes] = None,
    header: Optional[NaomiRom] = None,
    verbose: bool = False,
) -> FileBytes:
    ...


def add_or_update_trojan(
    data: Union[bytes, FileBytes],
    trojan: bytes,
    debug: int,
    options: int,
    datachunk: Optional[bytes] = None,
    header: Optional[NaomiRom] = None,
    verbose: bool = False,
) -> Union[bytes, FileBytes]:
    # Note that if an external header is supplied, it will be modified to match the updated
    # data.
    if isinstance(data, FileBytes):
        data = data.clone()
    if header is None:
        header = NaomiRom(data)

    # Grab a safe-to-mutate cop of the trojan, get its current config.
    executable = header.main_executable
    exe = trojan[:]
    _, location, _, _ = get_config(exe)

    for sec in executable.sections:
        if sec.load_address == location:
            # Grab the old entrypoint from the existing modification since the ROM header
            # entrypoint will be the old trojan EXE.
            entrypoint, _, _, _ = get_config(data, start=sec.offset, end=sec.offset + sec.length)
            exe = patch_bytesequence(exe, 0xAA, struct.pack("<I", entrypoint))
            if datachunk:
                exe = patch_bytesequence(exe, 0xBB, datachunk)
            exe = patch_bytesequence(exe, 0xCF, struct.pack("<I", options))
            exe = patch_bytesequence(exe, 0xDD, struct.pack("<I", debug))

            # We can reuse this section, but first we need to get rid of the old patch.
            if sec.offset + sec.length == len(data):
                # We can just stick the new file right on top of where the old was.
                if verbose:
                    print("Overwriting old section in existing ROM section.")

                # Cut off the old section, add our new section, make sure the length is correct.
                if isinstance(data, bytes):
                    data = data[:sec.offset] + exe
                elif isinstance(data, FileBytes):
                    data.truncate(sec.offset)
                    data.append(exe)
                else:
                    raise Exception("Logic error!")
                sec.length = len(exe)
            else:
                # It is somewhere in the middle of an executable, zero it out and
                # then add this section to the end of the ROM.
                if verbose:
                    print("Zeroing out old section in existing ROM section and attaching new section to the end of the file.")

                # Patch the executable with the correct settings and entrypoint.
                data = change(data, b"\0" * sec.length, sec.offset)

                # Repoint the section at the new section
                sec.offset = len(data)
                sec.length = len(exe)
                sec.load_address = location

                # Add the section to the end of the ROM.
                data = data + exe
            break
    else:
        if len(executable.sections) >= 8:
            raise NaomiSettingsPatcherException("ROM already has the maximum number of init sections!")

        # Add a new section to the end of the rom for this binary data.
        if verbose:
            print("Attaching section to a new ROM section at the end of the file.")

        executable.sections.append(
            NaomiRomSection(
                offset=len(data),
                load_address=location,
                length=len(exe),
            )
        )

        # Patch the executable with the correct settings and entrypoint.
        exe = patch_bytesequence(exe, 0xAA, struct.pack("<I", executable.entrypoint))
        if datachunk:
            exe = patch_bytesequence(exe, 0xBB, datachunk)
        exe = patch_bytesequence(exe, 0xCF, struct.pack("<I", options))
        exe = patch_bytesequence(exe, 0xDD, struct.pack("<I", debug))
        data = data + exe

    # Generate new header and attach executable to end of data.
    executable.entrypoint = location
    header.main_executable = executable

    # Now, return the updated ROM with the new header and appended data.
    if isinstance(data, bytes):
        return header.data + data[header.HEADER_LENGTH:]
    elif isinstance(data, FileBytes):
        data[:header.HEADER_LENGTH] = header.data
        return data
    else:
        raise Exception("Logic error!")


class NaomiSettingsPatcherException(Exception):
    pass


class NaomiSettingsDate:
    def __init__(self, year: int, month: int, day: int) -> None:
        self.year = year
        self.month = month
        self.day = day


class NaomiSettingsInfo:
    def __init__(self, debug: bool, date: Tuple[int, int, int]) -> None:
        self.enable_debugging = debug
        self.date = NaomiSettingsDate(date[0], date[1], date[2])


BytesLike = TypeVar("BytesLike", bound=Union[bytes, FileBytes])


class NaomiSettingsPatcher(Generic[BytesLike]):
    SRAM_LOCATION = 0x200000
    SRAM_SIZE = 32768

    EEPROM_SIZE = 128

    MAX_TROJAN_SIZE = 512 * 1024

    def __init__(self, rom: BytesLike, trojan: Optional[bytes]) -> None:
        self.__data: Union[bytes, FileBytes]
        if isinstance(rom, FileBytes):
            self.__data = rom.clone()
        elif isinstance(rom, bytes):
            self.__data = rom
        else:
            raise Exception("Logic error!")
        self.__trojan: Optional[bytes] = trojan
        if trojan is not None and len(trojan) > NaomiSettingsPatcher.MAX_TROJAN_SIZE:
            raise Exception("Logic error! Adjust the max trojan size to match the compiled tojan!")
        self.__rom: Optional[NaomiRom] = None
        self.__has_sram: Optional[bool] = None
        self.__has_eeprom: Optional[bool] = None

    @property
    def data(self) -> BytesLike:
        return cast(BytesLike, self.__data)

    @property
    def serial(self) -> bytes:
        # Parse the ROM header so we can grab the game serial code.
        naomi = self.__rom or NaomiRom(self.__data)
        self.__rom = naomi

        return naomi.serial

    @property
    def rom(self) -> NaomiRom:
        # Grab the entire rom as a parsed structure.
        naomi = self.__rom or NaomiRom(self.__data)
        self.__rom = naomi

        return naomi

    @property
    def has_eeprom(self) -> bool:
        if self.__has_eeprom is None:
            self.__has_eeprom = (self.eeprom_info is not None)
        return self.__has_eeprom

    @property
    def eeprom_info(self) -> Optional[NaomiSettingsInfo]:
        # Parse the ROM header so we can narrow our search.
        naomi = self.__rom or NaomiRom(self.__data)
        self.__rom = naomi

        # Only look at main executables.
        executable = naomi.main_executable
        for sec in executable.sections:
            # Constrain the search to the section that we jump to, since that will always
            # be where our trojan is.
            if executable.entrypoint >= sec.load_address and executable.entrypoint < (sec.load_address + sec.length):
                if sec.length > NaomiSettingsPatcher.MAX_TROJAN_SIZE:
                    # This can't possibly be the trojan, just skip it for speed.
                    continue
                try:
                    # Grab the old entrypoint from the existing modification since the ROM header
                    # entrypoint will be the old trojan EXE.
                    _, _, debug, date = get_config(self.__data, start=sec.offset, end=sec.offset + sec.length)

                    return NaomiSettingsInfo(debug, date)
                except Exception:
                    continue

        return None

    def get_eeprom(self) -> Optional[bytes]:
        # Parse the ROM header so we can narrow our search.
        naomi = self.__rom or NaomiRom(self.__data)
        self.__rom = naomi

        # Only look at main executables.
        executable = naomi.main_executable
        for sec in executable.sections:
            # Constrain the search to the section that we jump to, since that will always
            # be where our trojan is.
            if executable.entrypoint >= sec.load_address and executable.entrypoint < (sec.load_address + sec.length):
                if sec.length > NaomiSettingsPatcher.MAX_TROJAN_SIZE:
                    # This can't possibly be the trojan, just skip it for speed.
                    continue
                try:
                    # Grab the old entrypoint from the existing modification since the ROM header
                    # entrypoint will be the old trojan EXE.
                    get_config(self.__data, start=sec.offset, end=sec.offset + sec.length)

                    # Returns the requested EEPRom that should be written prior to the game starting.
                    for i in range(sec.offset, sec.offset + sec.length - (self.EEPROM_SIZE - 1)):
                        if NaomiEEPRom.validate(self.__data[i:(i + self.EEPROM_SIZE)], serial=naomi.serial):
                            self.__has_eeprom = True
                            return self.__data[i:(i + self.EEPROM_SIZE)]
                except Exception:
                    pass

        # Couldn't find a section that matched.
        self.__has_eeprom = False
        return None

    def put_eeprom(self, eeprom: bytes, *, enable_debugging: bool = False, verbose: bool = False) -> None:
        # First, parse the ROM we were given.
        naomi = self.__rom or NaomiRom(self.__data)

        # Now make sure the EEPROM is valid.
        if len(eeprom) == self.EEPROM_SIZE:
            # First, we need to modify the settings trojan with this ROM's load address and
            # the EEPROM we want to add. Make sure the EEPRom we were given is valid.
            if not NaomiEEPRom.validate(eeprom, serial=naomi.serial):
                raise NaomiSettingsPatcherException("EEPROM is incorrectly formed!")
            if naomi.serial != eeprom[3:7] or naomi.serial != eeprom[21:25]:
                raise NaomiSettingsPatcherException("EEPROM is not for this game!")

        else:
            raise NaomiSettingsPatcherException("Invalid EEPROM size to attach to a Naomi ROM!")

        # Now we need to add an EXE init section to the ROM.
        if self.__trojan is None or not self.__trojan:
            raise NaomiSettingsPatcherException("Cannot have an empty trojan when attaching EEPROM settings!")

        # Patch the trojan onto the ROM, updating the settings in the trojan accordingly.
        self.__data = add_or_update_trojan(
            self.__data,
            self.__trojan,
            1 if enable_debugging else 0,
            0,
            datachunk=eeprom,
            header=naomi,
            verbose=verbose,
        )

        # Also, write back the new ROM.
        self.__rom = naomi
        self.__has_eeprom = True

    @property
    def has_sram(self) -> bool:
        if self.__has_sram is None:
            # Need to see if there is an attached SRAM section in this exe.
            naomi = self.__rom or NaomiRom(self.__data)
            self.__rom = naomi

            for section in naomi.main_executable.sections:
                if section.load_address == self.SRAM_LOCATION and section.length == self.SRAM_SIZE:
                    self.__has_sram = True
                    break
            else:
                self.__has_sram = False
        return self.__has_sram

    def get_sram(self) -> Optional[bytes]:
        # Parse the ROM header so we can narrow our search.
        naomi = self.__rom or NaomiRom(self.__data)
        self.__rom = naomi

        # Only look at main executables.
        executable = naomi.main_executable
        for sec in executable.sections:
            # Check and see if it is a SRAM chunk.
            if sec.load_address == self.SRAM_LOCATION and sec.length == self.SRAM_SIZE:
                # It is!
                self.__has_sram = True
                return self.__data[sec.offset:(sec.offset + sec.length)]

        # Couldn't find a section that matched.
        self.__has_sram = False
        return None

    def put_sram(self, sram: bytes, *, verbose: bool = False) -> None:
        # First, parse the ROM we were given.
        naomi = self.__rom or NaomiRom(self.__data)

        # Now make sure the SRAM is valid.
        if len(sram) != self.SRAM_SIZE:
            raise NaomiSettingsPatcherException("Invalid SRAM size to attach to a Naomi ROM!")

        # Patch the section directly onto the ROM.
        self.__has_sram = True
        self.__data = add_or_update_section(self.__data, self.SRAM_LOCATION, sram, header=naomi, verbose=verbose)

        # Also, write back the new ROM.
        self.__rom = naomi



#! /usr/bin/env python3
from typing import Union, overload

from arcadeutils import FileBytes
from naomi.rom import NaomiRom


@overload
def _patch_rom(data: bytes, search: bytes, replace: bytes) -> bytes:
    ...


@overload
def _patch_rom(data: FileBytes, search: bytes, replace: bytes) -> FileBytes:
    ...


def _patch_rom(data: Union[bytes, FileBytes], search: bytes, replace: bytes) -> Union[bytes, FileBytes]:
    rom = NaomiRom(data)
    if not rom.valid:
        raise Exception("ROM file does not appear to be a Naomi netboot ROM!")

    # Now, find the location of our magic string in the main executable.
    exec_data = rom.main_executable
    patch_location = None
    search_len = len(search)

    # Look through all copied sections.
    for section in exec_data.sections:
        if isinstance(data, bytes):
            for rloc in range(section.length - (search_len - 1)):
                loc = rloc + section.offset
                if data[loc:(loc + search_len)] == search:
                    patch_location = loc
                    break
            if patch_location is not None:
                break
        elif isinstance(data, FileBytes):
            patch_location = data.search(search, start=section.offset, end=section.offset + section.length)
            if patch_location is not None:
                break

    if patch_location is None:
        raise Exception("ROM file does not have a suitable spot for the patch!")

    # Now, generate a patch with this updated data overlaid on the original rom
    if isinstance(data, bytes):
        return data[:patch_location] + replace + data[(patch_location + len(replace)):]
    elif isinstance(data, FileBytes):
        data = data.clone()
        data[patch_location:(patch_location + len(replace))] = replace
        return data
    else:
        raise Exception("Logic error!")


@overload
def force_freeplay(data: bytes) -> bytes:
    ...


@overload
def force_freeplay(data: FileBytes) -> FileBytes:
    ...


def force_freeplay(data: Union[bytes, FileBytes]) -> Union[bytes, FileBytes]:
    try:
        return _patch_rom(data, bytes([0x42, 0x84, 0xEC, 0x31, 0x0C, 0x60, 0x04, 0x1E, 0x43, 0x84]), bytes([0x1A, 0xE0]))
    except Exception:
        pass

    try:
        return _patch_rom(data, bytes([0x42, 0x84, 0x5c, 0x31, 0x1c, 0x7e, 0x0c, 0x60, 0x04, 0x15]), bytes([0x1A, 0xE0]))
    except Exception:
        pass

    try:
        return _patch_rom(data, bytes([0x42, 0x84, 0x5c, 0x31, 0x00, 0xeb, 0x0c, 0x60, 0x04, 0x15]), bytes([0x1A, 0xE0]))
    except Exception:
        pass

    raise Exception("Could not find any pattern to patch!")


@overload
def force_no_attract_sound(data: bytes) -> FileBytes:
    ...


@overload
def force_no_attract_sound(data: FileBytes) -> FileBytes:
    ...


def force_no_attract_sound(data: Union[bytes, FileBytes]) -> Union[bytes, FileBytes]:
    try:
        return _patch_rom(data, bytes([0x40, 0x63, 0x12, 0xe2, 0xec, 0x32, 0x3c, 0x63, 0x09, 0x43]), bytes([0x00, 0xe3]))
    except Exception:
        pass

    try:
        return _patch_rom(data, bytes([0x59, 0x23, 0x31, 0x1e, 0xe3, 0x63, 0x10, 0x73, 0x40, 0x62]), bytes([0x00, 0xe3]))
    except Exception:
        pass

    raise Exception("Could not find any pattern to patch!")



from naomi.settings.settings import (
    NaomiSettingsWrapper,
    NaomiSettingsManager,
    get_default_settings_directory,
)
from settings.settings import (
    SettingsParseException,
    SettingsSaveException,
    JSONParseException,
    ReadOnlyCondition,
    DefaultCondition,
    DefaultConditionGroup,
    Setting,
    Settings,
    SettingsConfig,
)

try:
    from naomi.settings.editor import NaomiSettingsEditor
except ModuleNotFoundError:
    # DragonCurses not installed, provide a dummy stub.
    class NaomiSettingsEditor:  # type: ignore
        def __init__(self, settings: NaomiSettingsWrapper, enable_unicode: bool = True) -> None:
            self.settings = settings

        def run(self) -> bool:
            return False


__all__ = [
    "NaomiSettingsWrapper",
    "NaomiSettingsManager",
    "NaomiSettingsEditor",
    "get_default_settings_directory",
    "SettingsParseException",
    "SettingsSaveException",
    "JSONParseException",
    "ReadOnlyCondition",
    "DefaultCondition",
    "DefaultConditionGroup",
    "Setting",
    "Settings",
    "SettingsConfig",
]



import os
from setuptools import setup


VERSION = "0.4.0"


if 'FULL_INSTALLATION' in os.environ:
    # We want to install this entire repo as an installation, so that we can
    # use it to run a netboot server.
    setup(
        name='netboot',
        version=VERSION,
        description='Code and utilities for netbooting a Naomi/Triforce/Chihiro, including a full web and netboot server.',
        author='DragonMinded',
        license='Public Domain',
        packages=[
            # Core packages.
            'netdimm',
            'naomi',
            'naomi.settings',
            'settings',
            # Include default trojan.
            'homebrew.settingstrojan',
            # Web server package.
            'netboot',
            'netboot.web',
            'netboot.web.static',
            'netboot.web.templates',
        ],
        install_requires=[
            req for req in open('requirements.txt').read().split('\n') if len(req) > 0
        ],
        package_data={
            # Make sure to actually include the trojan data.
            "homebrew.settingstrojan": ["settingstrojan.bin"],
            "netboot.web.static": ["*.js", "*.css"],
            "netboot.web.templates": ["*.html"],
        },
        python_requires=">=3.6",
    )
elif 'NAOMI_INSTALLATION' in os.environ:
    # We want to install only the naomi parts of this repo useful as a third-party
    # package so that other code can depend on us.
    with open(os.path.join("naomi", "README.md"), "r", encoding="utf-8") as fh:
        long_description = fh.read()
    long_description += os.linesep
    with open(os.path.join("naomi", "settings", "README.md"), "r", encoding="utf-8") as fh:
        long_description += fh.read()
    long_description += os.linesep
    with open(os.path.join("naomi", "settings", "definitions", "README.md"), "r", encoding="utf-8") as fh:
        long_description += fh.read()

    setup(
        name='naomiutils',
        version=VERSION,
        description='Code libraries for working with Naomi ROMs and EEPROMs.',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='DragonMinded',
        author_email='dragonminded@dragonminded.com',
        license='Public Domain',
        url='https://github.com/DragonMinded/netboot',
        packages=[
            # Package for 3rd party.
            'naomi',
            'naomi.settings',
            'settings',
            # Include settings definitions.
            'naomi.settings.definitions',
            # Include default trojan.
            'homebrew.settingstrojan',
        ],
        package_data={
            # Make sure mypy sees us as typed.
            "naomi": ["py.typed", "README.md"],
            "naomi.settings": ["py.typed", "README.md"],
            "settings": ["py.typed", "README.md"],
            # Make sure to include all existing settings.
            "naomi.settings.definitions": ["*.settings", "README.md"],
            # Make sure to actually include the trojan data.
            "homebrew.settingstrojan": ["settingstrojan.bin"],
        },
        install_requires=[
            'arcadeutils',
            'dragoncurses',
        ],
        python_requires=">=3.6",
    )
elif 'NETDIMM_INSTALLATION' in os.environ:
    # We want to install only the netdimm parts of this repo useful as a third-party
    # package so that other code can depend on us.
    with open(os.path.join("netdimm", "README.md"), "r", encoding="utf-8") as fh:
        long_description = fh.read()

    setup(
        name='netdimmutils',
        version=VERSION,
        description='Code libraries for working with a SEGA Net Dimm',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='DragonMinded',
        author_email='dragonminded@dragonminded.com',
        license='Public Domain',
        url='https://github.com/DragonMinded/netboot',
        packages=[
            # Package for 3rd party.
            'netdimm',
        ],
        package_data={
            # Make sure mypy sees us as typed.
            "netdimm": ["py.typed", "README.md"],
        },
        install_requires=[
            'pycryptodome',
            'arcadeutils',
            'naomiutils',
        ],
        python_requires=">=3.6",
    )
else:
    # We want to install the 3rdparty parts of this repo (netdimm and naomi) together
    # as a set of packages to depend on.
    setup(
        name='netbootutils',
        version=VERSION,
        description='Code and utilities for netbooting a Naomi/Triforce/Chihiro.',
        author='DragonMinded',
        license='Public Domain',
        packages=[
            # Core packages.
            'netdimm',
            'naomi',
            'naomi.settings',
            'settings',
            # Include settings definitions.
            'naomi.settings.definitions',
            # Include default trojan.
            'homebrew.settingstrojan',
        ],
        install_requires=[
            'pycryptodome',
            'arcadeutils',
            'dragoncurses',
        ],
        package_data={
            # Make sure mypy sees us as typed.
            "netdimm": ["py.typed", "README.md"],
            "naomi": ["py.typed", "README.md"],
            "naomi.settings": ["py.typed", "README.md"],
            "settings": ["py.typed", "README.md"],
            # Make sure to include all existing settings.
            "naomi.settings.definitions": ["*.settings", "README.md"],
            # Make sure to actually include the trojan data.
            "homebrew.settingstrojan": ["settingstrojan.bin"],
        },
    )
